#ifndef __CPHIDGET
#define __CPHIDGET

typedef struct _CPhidget CPhidget, *CPhidgetHandle;

#ifndef EXTERNALPROTO
#include "cphidgetattr.h"
#include "cphidgetconstantsinternal.h"
#include "cthread.h"

//This is for a socket - an one of these can be referenced by several phidgets!
typedef struct __CPhidgetSocketClient {
	int socket;
	char *port;
	char *address;
	//char *serverID;
	void *pdcs;
	int status;
	CThread_mutex_t lock; /* protects server status */
	CThread_mutex_t pdc_lock; /* protects synchronous pdc functions */
} CPhidgetSocketClient, *CPhidgetSocketClientHandle;

typedef struct __CPhidgetSocketClientList
{
	struct __CPhidgetSocketClientList *next;
	CPhidgetSocketClientHandle client;
} CPhidgetSocketClientList, *CPhidgetSocketClientListHandle;

typedef struct __CPhidgetRemote {
	CPhidgetSocketClientHandle server;	//can be NULL if not yet connected to a server
	char *requested_port;		//only !NULL if not zeroconf
	char *requested_address;	//only !NULL if not zeroconf
	char *requested_serverID;	//only !NULL if zeroconf
	char *password;				//can be NULL - will be used if needed for connecting to remote Phidgets
	int listen_id;
	int mdns;					//true if mdns, false if not
	char *zeroconf_name;		//for use before a connection is active
	char *zeroconf_domain;		//for use before a connection is active
	char *zeroconf_type;		//for use before a connection is active
	char *zeroconf_host;
	char *zeroconf_port;
	char *zeroconf_server_id;	//for use before a connection is active
	int zeroconf_auth;			//for use before a connection is active
} CPhidgetRemote, *CPhidgetRemoteHandle;

typedef struct __CPhidgetRemoteList
{
	struct __CPhidgetRemoteList *next;
	CPhidgetRemoteHandle networkInfo;
} CPhidgetRemoteList, *CPhidgetRemoteListHandle;

struct _CPhidget {
	CPhidgetRemoteHandle networkInfo; //NULL if local, !NULL if remote
	int(CCONV *fptrError)(CPhidgetHandle , void *, int, const char *);
	void *fptrErrorptr;
	int(CCONV *fptrConnect)(CPhidgetHandle , void *);
	void *fptrConnectptr; 
	int(CCONV *fptrDisconnect)(CPhidgetHandle , void *);
	void *fptrDisconnectptr;
	CThread_mutex_t lock; /* protects status */
	int status;
	CThread_mutex_t openCloseLock; /* protects status */
	int keyCount; //counts key during network open
	CThread_mutex_t writelock; /* protects write - exclusive */
	CThread readThread;
	CThread writeThread;
	int specificDevice;
	unsigned short deviceID;
	unsigned short deviceIDSpec;
	unsigned short outputReportByteLength;
	unsigned short inputReportByteLength;
	HANDLE deviceHandle;
#ifdef _WINDOWS
	OVERLAPPED asyncRead;
	BOOL readPending;
	EVENT closeReadEvent;
	OVERLAPPED asyncWrite;
	EVENT closeWriteEvent;
	unsigned char inbuf[9];
#endif
	const char *deviceType;
	int deviceVersion;
	unsigned short ProductID;
	unsigned short VendorID;
	int serialNumber;
	int(CCONV *fptrInit)(CPhidgetHandle);
	int(CCONV *fptrEvents)(CPhidgetHandle);
	int(CCONV *fptrData)(CPhidgetHandle, unsigned char *buffer, int length);
	int(CCONV *fptrGetPacket)(CPhidgetHandle, unsigned char *buffer,
	  unsigned int *length);
	unsigned char lastReadPacket[8];
	//char	ascii_representation[89];
	char label[11];
	unsigned char awdc_enabled;
	void *dnsServiceRef;
	void *CPhidgetFHandle;
	CThread_mutex_t outputLock; /* device-specific code responsible */
	EVENT writeAvailableEvent; /* device-specific code sets w/o OLL held */
	EVENT writtenEvent; /* device-specific code clears w/OLL held */
	int writeStopFlag; /* set when closing */
	int(CCONV *fptrAttach)(CPhidgetHandle , void *);
	void *fptrAttachptr; 
	int(CCONV *fptrDetach)(CPhidgetHandle , void *);
	void *fptrDetachptr;
	CPhidgetAttr attr;
};

struct _CPhidgetList
{
	struct _CPhidgetList *next;
	CPhidgetHandle phid;
} typedef CPhidgetList, *CPhidgetListHandle;

typedef struct {
	int pdd_did;
	int pdd_vid;
	int pdd_pid;
	CPhidgetAttr pdd_attr;
	int pdd_iid;
} CPhidgetDeviceDef;

extern const char LibraryVersion[];
extern const char *Phid_DeviceName[PHIDGET_DEVICE_TYPE_COUNT];
extern const char *Phid_DeviceSpecificName[PHIDGET_DEVICE_COUNT];
extern const char Phid_Device_Def_WorkAroundA[PHIDGET_DEVICE_COUNT][2];
extern const char Phid_UnknownErrorDescription[];
extern const char *Phid_ErrorDescriptions[PHIDGET_ERROR_CODE_COUNT];

extern CPhidgetList *ActiveDevices;
extern CPhidgetList *AttachedDevices;
extern int phidgetLocksInitialized;
extern CThread_mutex_t activeDevicesLock, attachedDevicesLock;

extern const CPhidgetDeviceDef Phid_Device_Def[];

int CPhidget_read(CPhidgetHandle phid);
int CPhidget_write(CPhidgetHandle phid);
//int CPhidget_isAttached(CPhidgetHandle phid);
int CPhidget_statusFlagIsSet(CPhidgetHandle phid, int flag);
int CPhidget_setStatusFlag(CPhidgetHandle phid, int flag);
int CPhidget_clearStatusFlag(CPhidgetHandle phid, int flag);
//void CPhidget_setStatus(CPhidgetHandle phid, int v);
//int CPhidget_setStatus(CPhidgetHandle phid, int v);
char translate_bool_to_ascii(char value);
const char *CPhidget_strerror(int error);
void throw_error_event(CPhidgetHandle phid, const char *error, int errcode);
int attachActiveDevice(CPhidgetHandle activeDevice, CPhidgetHandle attachedDevice);

PHIDGET21_API int CCONV CPhidget_areEqual(void *arg1, void *arg2);
PHIDGET21_API int CCONV CPhidget_areExtraEqual(void *arg1, void *arg2);
PHIDGET21_API int CCONV CPhidgetHandle_areEqual(void *arg1, void *arg2);
PHIDGET21_API void CCONV CPhidget_free(void *arg);
PHIDGET21_API int CCONV phidget_type_to_id(const char *name);
PHIDGET21_API int CCONV CPhidget_create(CPhidgetHandle *phid);

PHIDGET21_API int CCONV CPhidget_calibrate_gainoffset(CPhidgetHandle phid, int Index, unsigned short offset, unsigned long gain);

#endif
#include "cphidgetmacros.h"

PHIDGET21_API int CCONV CPhidget_open(CPhidgetHandle phid, int serialNumber);
PHIDGET21_API int CCONV CPhidget_close(CPhidgetHandle phid);
PHIDGET21_API int CCONV CPhidget_delete(CPhidgetHandle phid);
PHIDGET21_API int CCONV CPhidget_set_OnDetach_Handler(CPhidgetHandle, int(CCONV *fptr)(CPhidgetHandle, void *), void *userPtr);
PHIDGET21_API int CCONV CPhidget_set_OnAttach_Handler(CPhidgetHandle, int(CCONV *fptr)(CPhidgetHandle, void *), void *userPtr);
PHIDGET21_API int CCONV CPhidget_set_OnServerConnect_Handler(CPhidgetHandle phid, int (CCONV *fptr)(CPhidgetHandle phid, void *userPtr), void *userPtr);
PHIDGET21_API int CCONV CPhidget_set_OnServerDisconnect_Handler(CPhidgetHandle phid, int (CCONV *fptr)(CPhidgetHandle phid, void *userPtr), void *userPtr);
PHIDGET21_API int CCONV CPhidget_set_OnError_Handler(CPhidgetHandle phid, int(CCONV *fptr)(CPhidgetHandle, void *, int, const char *), void *userPtr);
PHIDGET21_API int CCONV CPhidget_getDeviceName(CPhidgetHandle phid,
    const char **buffer);
PHIDGET21_API int CCONV CPhidget_getSerialNumber(CPhidgetHandle phid,
    int *serialNumber);
PHIDGET21_API int CCONV CPhidget_getDeviceVersion(CPhidgetHandle phid,
    int *devVer);
PHIDGET21_API int CCONV CPhidget_getDeviceStatus(CPhidgetHandle phid,
    int *status);
PHIDGET21_API int CCONV CPhidget_getLibraryVersion(const char **buffer);
PHIDGET21_API int CCONV CPhidget_getDeviceType(CPhidgetHandle phid,
    const char **buffer);
PHIDGET21_API int CCONV CPhidget_getDeviceLabel(CPhidgetHandle phid, const char **buffer);
PHIDGET21_API int CCONV CPhidget_setDeviceLabel(CPhidgetHandle phid, const char *buffer);
PHIDGET21_API int CCONV CPhidget_getErrorDescription(int ErrorCode, const char **buf);
PHIDGET21_API int CCONV CPhidget_waitForAttachment(CPhidgetHandle phid, int milliseconds);
PHIDGET21_API int CCONV CPhidget_getServerID(CPhidgetHandle phid, const char **serverID);
PHIDGET21_API int CCONV CPhidget_getServerAddress(CPhidgetHandle phid, const char **address, int *port);
PHIDGET21_API int CCONV CPhidget_getServerStatus(CPhidgetHandle phid, int *status);

#endif
