#include "stdafx.h"
#include "cphidgetgyroscope.h"
#include "math.h"
#include "cusb.h"
#include "csocket.h"
#include "cthread.h"

// === Internal Functions === //

//initAfterOpen - sets up the initial state of an object, reading in packets from the device if needed
//				  used during attach initialization - on every attach
CPHIDGETINIT(Gyroscope)
	int i = 0;

	TESTPTR(phid);

	//TODO:Setup max/min values

	//initialize triggers, set data arrays to unknown
	for (i = 0; i<phid->phid.attr.gyroscope.numAxis; i++)
	{
		phid->axis[i] = PUNK_DBL;
		phid->axisLastTrigger[i] = PUNK_DBL;
		phid->axisChangeTrigger[i] = 0;
	}

	//issue one read
	CPhidget_read((CPhidgetHandle)phid);

	return EPHIDGET_OK;
}

//dataInput - parses device packets
CPHIDGETDATA(Gyroscope)
	int i;
	double axis[GYRO_MAXAXES];

	if (length<0) return EPHIDGET_INVALIDARG;
	TESTPTR(phid);
	TESTPTR(buffer);

	ZEROMEM(axis, sizeof(axis));
	
	//Parse device packets - store data locally
	switch(phid->phid.deviceIDSpec)
	{
		case PHIDSPEC_GYRO_w_ACCELEROMETER:
			if (phid->phid.deviceVersion < 200)
			{
				i = ((unsigned short)buffer[0]+((unsigned short)buffer[1]<<8));
				axis[0] = round_double((((double)(i-32768)) / 196.605), 4);
				i = ((unsigned short)buffer[2]+((unsigned short)buffer[3]<<8));
				axis[1] = round_double((((double)(i-32768)) / 196.605), 4);
				i = ((unsigned short)buffer[4]+((unsigned short)buffer[5]<<8));
				axis[2] = round_double((((double)(i-32768)) / 196.605), 4);
			}
			break;
		default:
			break;
	}

	//Make sure values are within defined range, and store to structure
	for (i = 0; i<phid->phid.attr.accelerometer.numAxis; i++)
	{
		phid->axis[i] = axis[i];
	}

	//send out any events that exceed or match the trigger
	for (i = 0; i<phid->phid.attr.gyroscope.numAxis; i++)
	{
		if (fabs(phid->axis[i] - phid->axisLastTrigger[i]) >= phid->axisChangeTrigger[i]
			|| phid->axisLastTrigger[i] == PUNK_DBL)
		{
			FIRE(AngularRateChange, i, phid->axis[i]);
			phid->axisLastTrigger[i] = phid->axis[i];
		}
	}

	return EPHIDGET_OK;
}

//eventsAfterOpen - sends out an event for all valid data, used during attach initialization
CPHIDGETINITEVENTS(Gyroscope)
	int i = 0;

	for (i = 0; i<phid->phid.attr.gyroscope.numAxis; i++)
	{
		if(phid->axis[i] != PUNK_DBL)
			FIRE(AngularRateChange, i, phid->axis[i]);
	}

	return EPHIDGET_OK;
}

//getPacket - not used for Gyroscope
CGETPACKET(Gyroscope)
	return EPHIDGET_UNEXPECTED;
}

// === Exported Functions === //

//create and initialize a device structure
CCREATE(Gyroscope, PHIDDEF_GYROSCOPE)

//event setup functions
CFHANDLE(Gyroscope, AngularRateChange, int, double)

CGET(Gyroscope,NumAxis,int)
	TESTPTRS(phid,pVal) 
	TESTDEVICETYPE(PHIDDEF_GYROSCOPE)
	TESTATTACHED
	MASGN(phid.attr.gyroscope.numAxis)
}

CGETINDEX(Gyroscope,AngularRate,double)
	TESTPTRS(phid,pVal) 
	TESTDEVICETYPE(PHIDDEF_GYROSCOPE)
	TESTATTACHED
	TESTINDEX(phid.attr.gyroscope.numAxis)
	TESTMASGN(axis[Index], PUNK_DBL)
	MASGN(axis[Index])
}

CGETINDEX(Gyroscope,AngularRateChangeTrigger,double)
	TESTPTRS(phid,pVal) 
	TESTDEVICETYPE(PHIDDEF_GYROSCOPE)
	TESTATTACHED
	TESTINDEX(phid.attr.gyroscope.numAxis)
	MASGN(axisChangeTrigger[Index])
}
CSETINDEX(Gyroscope,AngularRateChangeTrigger,double)
	TESTPTR(phid) 
	TESTDEVICETYPE(PHIDDEF_GYROSCOPE)
	TESTATTACHED
	TESTINDEX(phid.attr.gyroscope.numAxis)
	if(phid->phid.networkInfo)
		ADDNETWORKKEYINDEXED(Trigger, "%lf");
	else
		phid->axisChangeTrigger[Index] = newVal;
	return EPHIDGET_OK;
}

