#ifndef __CPHIDGETMANAGER
#define __CPHIDGETMANAGER

#include "cphidget.h"

DPHANDLE(Manager)

#ifndef EXTERNALPROTO

typedef enum
{
	PHIDGETMANAGER_INACTIVE = 1,
	PHIDGETMANAGER_ACTIVE
} CPhidgetManagerState;

typedef struct _CPhidgetManager
{
	CPhidgetRemoteHandle networkInfo;
	int(CCONV *fptrError)(CPhidgetManagerHandle , void *, int, const char *);
	void *fptrErrorptr;
	int(CCONV *fptrConnect)(CPhidgetManagerHandle , void *);
	void *fptrConnectptr; 
	int(CCONV *fptrDisconnect)(CPhidgetManagerHandle , void *);
	void *fptrDisconnectptr;
	CThread_mutex_t lock; /* protects status */
	int status;
	CPhidgetManagerState state;
	CPhidgetList *AttachedPhidgets;
	int (CCONV *fptrAttachChange)(CPhidgetHandle phid, void *userPtr);
	void *fptrAttachChangeptr;
	int (CCONV *fptrDetachChange)(CPhidgetHandle phid, void *userPtr);
	void *fptrDetachChangeptr;
} CPhidgetManager;

typedef struct _CPhidgetManagerList
{
	struct _CPhidgetManagerList *next;
	CPhidgetManager *phidm;
} CPhidgetManagerList, *CPhidgetManagerListHandle;

extern CPhidgetManagerListHandle localPhidgetManagers;
extern int ActivePhidgetManagers;

int CPhidgetManager_poll();
int CPhidgetAttachEvent(CPhidgetHandle phid);
int CPhidgetDetachEvent(CPhidgetHandle phid);

int CPhidgetManager_areEqual(void *arg1, void *arg2);
void CPhidgetManager_free(void *arg);
#endif

PHIDGET21_API int CCONV CPhidgetManager_create(CPhidgetManagerHandle *phidm);
PHIDGET21_API int CCONV CPhidgetManager_open(CPhidgetManagerHandle phidm);
PHIDGET21_API int CCONV CPhidgetManager_close(CPhidgetManagerHandle phidm);
PHIDGET21_API int CCONV CPhidgetManager_delete(CPhidgetManagerHandle phidm);
PHIDGET21_API int CCONV CPhidgetManager_set_OnAttach_Handler(CPhidgetManagerHandle phidm, int (CCONV *fptr)(CPhidgetHandle phid, void *userPtr), void *);
PHIDGET21_API int CCONV CPhidgetManager_set_OnDetach_Handler(CPhidgetManagerHandle phidm, int (CCONV *fptr)(CPhidgetHandle phid, void *userPtr), void *);
PHIDGET21_API int CCONV CPhidgetManager_getAttachedDevices(CPhidgetManagerHandle phidm, CPhidgetHandle *phidArray[], int *count);
PHIDGET21_API int CCONV CPhidgetManager_set_OnError_Handler(CPhidgetManagerHandle phidm, int(CCONV *fptr)(CPhidgetManagerHandle, void *, int, const char *), void *userPtr);
PHIDGET21_API int CCONV CPhidgetManager_set_OnServerConnect_Handler(CPhidgetManagerHandle phidm, int (CCONV *fptr)(CPhidgetManagerHandle phidm, void *userPtr), void *userPtr);
PHIDGET21_API int CCONV CPhidgetManager_set_OnServerDisconnect_Handler(CPhidgetManagerHandle phidm, int (CCONV *fptr)(CPhidgetManagerHandle phidm, void *userPtr), void *userPtr);
PHIDGET21_API int CCONV CPhidgetManager_getServerID(CPhidgetManagerHandle phidm, const char **serverID);
PHIDGET21_API int CCONV CPhidgetManager_getServerAddress(CPhidgetManagerHandle phidm, const char **address, int *port);
PHIDGET21_API int CCONV CPhidgetManager_getServerStatus(CPhidgetManagerHandle phidm, int *status);
#endif

